/* OTTO support procedures.  */

#include "otto_platform.h"

void microdelay (unsigned usecs);

/* Dummy buffer used for the final DMA fragment.   This fragment is
   used to terminate all DMA chains.   It will swallow all received
   packets that are too long and source all transmit packets that are too
   long. */
#define DUMMYLEN (1 << 8)	/* must be power of two > 2 and <= page size */
static unsigned char dummybuffer[2*DUMMYLEN];
/* allocate 2*length to allow for rounding to make physically contiguous */

/* initialize the device */
int otto_initialize (int unit, char *devicebase, struct otto_data *od,
	struct otto_xilinxdesc *xilinxcode,
	unsigned txcells, unsigned noackcells, char *reportbase)
{
	otto_sethwaddresses (devicebase, unit, od, reportbase);
	return (otto_fullreset (od, xilinxcode, txcells, noackcells));
}

int otto_fullreset (struct otto_data *od, struct otto_xilinxdesc *xilinxcode,
	unsigned txcells, unsigned noackcells)
{
	if (otto_loadxilinx (od, xilinxcode) == -1) {	/* load xilinxes */
		printf ("otto: xilinx load failed\n");
		return (-1);
	}
	return (otto_softreset (od, txcells, noackcells));
}

#define BASIC_RESETS (OTTO_CSR_RSTXILINX | OTTO_CSR_RSTALTERA | OTTO_CSR_RSTSUNI)
/* send a single bit to the UID serial ROM */
static uidromsendbit (struct otto_data *od, int bit)
{
	bit = (bit? OTTO_CSR_HOSTID_DI: 0);
	od->csr->resets = BASIC_RESETS | OTTO_CSR_HOSTID_CS | bit;
	wbflush ();
	microdelay (1);
	od->csr->resets =
		BASIC_RESETS | OTTO_CSR_HOSTID_CS | bit | OTTO_CSR_HOSTID_SK;
	wbflush ();
	microdelay (1);
}

/* check the UID ROM - put the UID in od->physaddress */
/* on entry to this routine, the device is in reset state */
static int uidok (struct otto_data *od)
{
	unsigned cksum;
	static unsigned char test[8] = {
		0xff, 0x00, 0x55, 0xaa, 0xff, 0x00, 0x55, 0xaa
	};
	unsigned char romcontents[66];
	unsigned char uida[8];
	int ok = 1;
	unsigned crc;
	int i;

	/* The UID serial ROM is arranged as follows:
		there are 33 interesting 16-bit words, which
		come out in little-endian order. (i.e. low address first)

		the bits within each 16 bit word are in bigendian order
		(i.e. high-numbered bits first).  Strange, but true.
	 */
	for (i = 0; i != 33; i++) {
		unsigned w = 0;
		int j;

		od->csr->resets = BASIC_RESETS;	/* turn off chip select */
		wbflush ();
		microdelay (1);
		od->csr->resets = BASIC_RESETS | OTTO_CSR_HOSTID_CS; /* turn on chip select */
		wbflush ();
		microdelay (1);

		uidromsendbit (od, 1);		/* send a 1 */
		uidromsendbit (od, 1);		/* send a 1 */
		uidromsendbit (od, 0);		/* send a 0 */

		/* send bits 5..0 of address */
		for (j = 5; j >= 0; j--) {
			uidromsendbit (od, (i >> j) & 1);	/* send bit */
		}

		for (j = 0; j != 16; j++) {
			unsigned x;
			w <<= 1;
			uidromsendbit (od, 0);		/* send a 0 */
			x = od->csr->resets;		/* read bit */
			if ((x & OTTO_CSR_HOSTID_DO) != 0) {
				w |= 1;
			}
		}
		romcontents[i*2] = w;
		romcontents[i*2+1] = w >> 8;

		od->csr->resets = BASIC_RESETS;	/* turn off chip select */
		wbflush ();
	}

	for (i = 0; i != 6; i++) {
		uida[i] = romcontents[i];
		od->physaddress.ether_addr_octet[i] = romcontents[i];
	}

	/* old/weak checksum */
	cksum = uida[0]*(1<<10) + uida[1]*(1<<2) +
	        uida[2]*(1<<9)  + uida[3]*(1<<1) +
	        uida[4]*(1<<8)  + uida[5]*(1<<0);
	cksum %= 0xffff;
	uida[6] = cksum >> 8;
	uida[7] = cksum;

	for (i = 0; i != 8; i++) {
		if (romcontents[i] != uida[i] ||
		    romcontents[i+8] != uida[7-i] ||
		    romcontents[i+16] != uida[i] ||
		    romcontents[i+24] != test[i]) {
			ok = 0;
		}
	}

        /* check new/strong CRC */
        crc = 0;
        for (i = 0; i != 64; i++) {
                unsigned temp = romcontents[i];
                int j;
                for (j = 0; j != 8; j++) {
                        unsigned xor = ((temp >> j) ^ crc) & 1;
			crc = (crc >> 1) ^ (xor? 0xa001 : 0);
                }
        }
        if (crc != ((romcontents[65] << 8) | romcontents[64])) {
                ok = 0;
        }
	return (ok);
}

int otto_softreset (struct otto_data *od, unsigned txcells, unsigned noackcells)
{
	volatile otto_uword32 *vcrings;
	int i;
	unsigned long phys;
	unsigned long phys1;

	otto_setreset (od);	/* enter reset state */
	microdelay (100);
	i = uidok (od);
	otto_unsetreset (od);	/* exit reset state */

	if (!i) {
		printf ("otto: UID ROM error\n");
		return (-1);
	}

	/* Initialize FIFOs */
	*od->cell = OTTO_CELL_FIFOFLAGS;
	wbflush ();
	*od->fifo = OTTO_FIFO_MAGIC;
	wbflush ();

	*od->cell = OTTO_CELL_USERIGHT;
	wbflush ();

	/* od->suni[8*SUNI_MC] = SUNI_MC_DLE; /* Put SUNI into loopback mode */
	od->suni[8*SUNI_MC] = 0x0;	 /* Put SUNI into normal mode */
	wbflush();

	/* turn off both leds */
	od->csr->leds = 0;
	wbflush ();


	otto_lockpioonly (od);	/* lock microsequencer into PIO-only mode */

	if (otto_sramtest (od) == -1) {		/* test the RAM */
		printf ("otto: RAM test failed\n");
		return (-1);
	}
	/* RAM is now zero */

	otto_setreportbase (od);

	otto_initfreelists (od);

	/* Construct final fragment */
	phys = OTTO_SVTOPHY (dummybuffer);
	/* ensure physically contiguous  (DUMMYLEN <= page size) */
	phys1 = OTTO_SVTOPHY (dummybuffer+DUMMYLEN);
	if (phys + DUMMYLEN != phys1) {
		phys = phys1;
	}
	od->descriptors[OTTO_FINALFRAG].word0 =
		OTTO_MAKE_FRAG0 (DUMMYLEN, OTTO_FINALFRAG);
	od->descriptors[OTTO_FINALFRAG].word1 = OTTO_MAKE_FRAG1 (phys);

	/* make a dummy packet descriptor */
   	od->descriptors[OTTO_EMPTYPKT].word0 = OTTO_MAKE_RXPKT (OTTO_DONTWRAP,
   		OTTO_EMPTYPKT, OTTO_TAG1, OTTO_FINALFRAG); 
	/* Construct empty ring-of-packets descriptor */
   	od->descriptors[OTTO_EMPTYRING].word0 = OTTO_MAKE_RING (OTTO_TAG0,
   		OTTO_EMPTYPKT);
	/* Construct full ring-of-packets descriptor */
   	od->descriptors[OTTO_FULLRING].word0 = OTTO_MAKE_RING (OTTO_TAG1,
   		OTTO_EMPTYPKT);


   	/* point all (as yet unused) receive VC's at full ring of
   	   packets descriptor */
	vcrings = &od->ram[OTTO_VCPOS+(1*OTTO_MAXVC)];
   	for (i = 0; i != OTTO_MAXVC; i++) {
		vcrings[i] = OTTO_MAKE_RXVCRING (OTTO_FULLRING);
   	}

	/* This makes VC 0 (= junk cells) not be in packet mode */
	otto_setrxvc (od, 0, 1, 0, 0, OTTO_FULLRING);

	/* construct the cell free lists */
	otto_buildcelllists (od, txcells, noackcells);

	/* initialize the SUNI chip */
	otto_initsuni (od);

	od->csr->interrupts = OTTO_CSR_DMAENABLE; /* enable DMA interrupts */

	otto_unlockpioonly (od);
	/* device can now send and receive packets */

	return (0);
}

/* initialize addresses and sizes in od */
void otto_sethwaddresses (char *devicebase, int unit, struct otto_data *od,
	char *reportbase)
{
	od->slotaddress = devicebase;
	od->rom = (volatile otto_uword32 *) (devicebase + OTTO_ROM);
	od->ram = (volatile otto_uword32 *) (devicebase + OTTO_RAM);
	od->descriptors = (volatile struct otto_descriptor *) &od->ram[OTTO_DESCPOS];
	od->ram_writelo = (volatile otto_uword32 *) (devicebase + OTTO_RAMB);
	od->ram_writehi = (volatile otto_uword32 *) (devicebase + OTTO_RAMA);
	od->csr = (volatile struct otto_csr *) (devicebase + OTTO_CSR);
	od->dma = (volatile struct otto_dma *) (devicebase + OTTO_DMA);
	od->fifo = (volatile otto_uword32 *) (devicebase + OTTO_FIFO);
	od->cell = (volatile otto_uword32 *) (devicebase + OTTO_CELL);
	od->suni = (volatile otto_uword32 *) (devicebase + OTTO_SUNI);
	od->exec = (volatile otto_uword32 *) (devicebase + OTTO_EXEC);
	od->txrequest = (volatile otto_uword32 *) (devicebase + OTTO_REQ);
	od->activate = (volatile otto_uword32 *) (devicebase + OTTO_ACTIVATE);
	od->ackfifoptrs =
		(volatile otto_uword32 *) (devicebase + OTTO_ACKFIFOPTRS);

	od->rom_intack = (volatile otto_uword32 *)
		(devicebase + OTTO_ROM + OTTO_INTACK);
	od->ram_intack = (volatile otto_uword32 *)
		(devicebase + OTTO_RAM + OTTO_INTACK);
	od->descriptors_intack = (volatile struct otto_descriptor *) &od->ram_intack[OTTO_DESCPOS];
	od->ram_writelo_intack = (volatile otto_uword32 *)
		(devicebase + OTTO_RAMB + OTTO_INTACK);
	od->ram_writehi_intack = (volatile otto_uword32 *)
		(devicebase + OTTO_RAMA + OTTO_INTACK);
	od->csr_intack = (volatile struct otto_csr *)
		(devicebase + OTTO_CSR + OTTO_INTACK);
	od->dma_intack = (volatile struct otto_dma *)
		(devicebase + OTTO_DMA + OTTO_INTACK);
	od->fifo_intack = (volatile otto_uword32 *)
		(devicebase + OTTO_FIFO + OTTO_INTACK);
	od->cell_intack = (volatile otto_uword32 *)
		(devicebase + OTTO_CELL + OTTO_INTACK);
	od->suni_intack = (volatile otto_uword32 *)
		(devicebase + OTTO_SUNI + OTTO_INTACK);
	od->exec_intack = (volatile otto_uword32 *)
		(devicebase + OTTO_EXEC + OTTO_INTACK);
	od->txrequest_intack = (volatile otto_uword32 *)
		(devicebase + OTTO_REQ + OTTO_INTACK);

	od->romsize = OTTO_ROMLEN;
	od->ramsize = OTTO_RAMLEN;
	od->ncellbuffers = OTTO_MAXCELLS;
	od->ndescriptors = OTTO_MAXDOUBLEDESC;
	od->reportringsize = OTTO_REP_RING_LEN_BYTES;

	od->reportbase = (volatile struct otto_rep *) reportbase;

	od->unit = unit;
	od->ucodeversion = 0;
	od->ucodecreated = 0;
	od->otto_sdh = 0; /* Default to Sonet Interface */
}

/* initialize the VC and descriptor free lists */
void otto_initfreelists (struct otto_data *od)
{
        unsigned i;
        unsigned j;

        /* Set up free list for ring descriptors.  Entry 0 in the array is
           never used */
	od->ringfreelist = 1;			/* head of free list */
	od->ringmissedlist = 0;			/* head of missed list */
	for (i = 1; i != OTTO_NRINGS; i++) {
		od->rings[i].next = i+1;
		od->rings[i].desc = OTTO_RINGSSTART +
			((OTTO_MAXPKTSPERRING+1)*i);
		od->rings[i].tag = OTTO_TAG0;
		od->rings[i].nextfree = 0;
		od->rings[i].lastused = 0;
		od->rings[i].missed = 0;
		bzero ((char *) od->rings[i].pktindex,
			sizeof (od->rings[i].pktindex));
	}
	od->rings[OTTO_NRINGS-1].next = 0;	/* end of free list */

        /* Set up free list for receive packet descriptors. 
           Entry 0 in the array is never used */
	od->rxpktfreelist = 1;			/* head of free list */
	od->rxpkts[0].m = (struct mbuf *) 0;
	for (i = 1; i != OTTO_NRXPKTS; i++) {
		od->rxpkts[i].next = i+1;
		od->rxpkts[i].desc = OTTO_RXPKTSTART +
			(OTTO_MAXFRAGSPERRXBUF*i);
		od->rxpkts[i].m = (struct mbuf *) 0;
	}
	od->rxpkts[OTTO_NRXPKTS-1].next = 0;	/* end of free list */


        /* Set up free list for transmit packet descriptors. 
           Entry 0 in the array is never used */
	od->txpktfreelist = 1;			/* head of free list */
	od->txpkts[0].m = (struct mbuf *) 0;
	j = 0;
	for (i = 1; i != OTTO_NTXPKTS; i++) {
		od->txpkts[i].next = i+1;
		od->txpkts[i].desc = OTTO_TXPKTSTART +
			((OTTO_MAXFRAGSPERTXBUF+1)*i);
		od->txpkts[i].m = (struct mbuf *) 0;
		do {
			od->txpkts[i].trailerv = &od->trailerspace[j];
			od->txpkts[i].trailerp =
				OTTO_SVTOPHY ((char *) od->txpkts[i].trailerv);
			j += OTTO_MAXTRAILERLEN;
		} while ((od->txpkts[i].trailerp + OTTO_MAXTRAILERLEN) !=
		    OTTO_SVTOPHY ((char *) (od->txpkts[i].trailerv +
		    	 OTTO_MAXTRAILERLEN)));
	}
	od->txpkts[OTTO_NTXPKTS-1].next = 0;	/* end of free list */
		    /* This list is not circular.  Entry 0 is
		       unused, and is being used as a list terminator */

	/* zero all transmit VC packet queues */
	for (i = 0; i != OTTO_MAXVC; i++) {
		od->vcs[i].txpkts = 0;
	}
}

/* after this call OTTO is in reset state */
void otto_setreset (struct otto_data *od)
{
	od->csr->resets =
		(OTTO_CSR_RSTXILINX|OTTO_CSR_RSTALTERA|OTTO_CSR_RSTSUNI);
	wbflush ();
}

/* after this call OTTO is no longer in reset state */
void otto_unsetreset (struct otto_data *od)
{
	od->csr->resets = 0;
	wbflush ();

	/* save the ucode version and timestamp written
	   by microcode when it comes out of reset state */
	od->ucodeversion = od->ram[0];
	od->ucodecreated = od->ram[1];
}

/* after this call, OTTO can do only PIO activity,
   and receive interrupts are turned on */
void otto_lockpioonly (struct otto_data *od)
{
	od->dma->status = (OTTO_DMA_STATUS_PIO_TRAN|OTTO_DMA_STATUS_PIO_RECV) &
		~OTTO_DMA_STATUS_PIO_UNLOCK;
	wbflush ();
}

/* after this call, OTTO can do non-PIO activity, and normal interrupts are
   enabled */
void otto_unlockpioonly (struct otto_data *od)
{
	wbflush ();
	od->dma->status = (OTTO_DMA_STATUS_PIO_TRAN|OTTO_DMA_STATUS_PIO_RECV) |
		OTTO_DMA_STATUS_PIO_UNLOCK;
	wbflush ();
}

/* after this call, OTTO can do non-PIO activity, normal interrupts are
   disabled, and timer inetrrupts are enabled  */
void otto_defer (struct otto_data *od)
{
	od->dma->status = (OTTO_DMA_STATUS_PIO_TICK) |
		OTTO_DMA_STATUS_PIO_UNLOCK;
	wbflush ();
}

/* load the Xilinxes on the OTTO */
int otto_loadxilinx (struct otto_data *od, struct otto_xilinxdesc *xilinxcode)
{
	int stepsize;
	int i;
	int len;
	int retry;
	unsigned char *x = (unsigned char *) 0;

	if (xilinxcode != (struct otto_xilinxdesc *) 0) {
		int j;
		int checksum;

		x = xilinxcode->data;
		stepsize = xilinxcode->stride;
		len = xilinxcode->len;

		checksum = 0;
		j = 0;
		for (i = 0; i != len; i++) {
			checksum += x[j];
			j += stepsize;
		}
		if (checksum != xilinxcode->checksum) {
			printf ("otto_loadxilinx: checksum failed\n");
			x = (unsigned char *) 0;
			return (-1);
		}
	}

	if (x == (unsigned char *) 0) {
		x = (unsigned char *) &od->rom[OTTO_ROM_XILINXCODE];
		stepsize = sizeof (otto_uword32);
		len = OTTO_ROM_XILINXCODESIZE;

		printf ("otto: ROM xilinxcode not there yet\n");
		return (-1);
	}

	for (retry = 0; retry != 5; retry++) {
		od->csr->resets = OTTO_CSR_RSTALTERA | OTTO_CSR_RSTSUNI;
		wbflush ();

		microdelay (100);

		od->csr->resets = OTTO_CSR_RSTXILINX | OTTO_CSR_RSTALTERA | 
			OTTO_CSR_RSTSUNI | OTTO_CSR_PROG;
		wbflush ();

		i = 1000;
		while ((od->csr->resets & OTTO_CSR_INIT) == 0 && i != 0) {
			i--;
		}
		if (i == 0) {
			cprintf (
				"otto_loadxilinx: failed: INIT low too long\n");
			microdelay (100);
			continue;
		}

		microdelay (100);

		od->csr->resets = OTTO_CSR_RSTALTERA | OTTO_CSR_RSTSUNI;

		i = 1000;
		while ((od->csr->resets & OTTO_CSR_INIT) != 0 && i != 0) {
			i--;
		}
		if (i == 0) {
			cprintf (
			    "otto_loadxilinx: failed: INIT high too long\n");
			microdelay (100);
			continue;
		}

		microdelay (100);

		for (i = 0; i != len; i++) {
			int byte = x[i];

			od->csr->loadxilinx = byte; byte >>= 1; wbflush ();
			od->csr->loadxilinx = byte; byte >>= 1; wbflush ();
			od->csr->loadxilinx = byte; byte >>= 1; wbflush ();
			od->csr->loadxilinx = byte; byte >>= 1; wbflush ();
			od->csr->loadxilinx = byte; byte >>= 1; wbflush ();
			od->csr->loadxilinx = byte; byte >>= 1; wbflush ();
			od->csr->loadxilinx = byte; byte >>= 1; wbflush ();
			od->csr->loadxilinx = byte; wbflush ();
		}

		if ((od->csr->resets & OTTO_CSR_NOT_DONE) != 0) {
			cprintf ("otto_loadxilinx: programming not complete\n");
			microdelay (100);
			continue;
		}

		/* success */
		return (0);
	}

	cprintf ("otto_loadxilinx: failed: too many attempts\n");
	return (-1);
}

void otto_uselink (struct otto_data *od, int link)
{
	if (link == 0) {
		*od->cell = OTTO_CELL_USERIGHT;
	} else {
		*od->cell = OTTO_CELL_USELEFT;
	}
	wbflush ();

	otto_wait_for_phaselock (od);
}

/* make sure we can write random values to the ram */
/* leave 0 in all the locations */
int otto_sramtest (struct otto_data *od)
{
	int i;
	int n = od->ramsize / sizeof (otto_uword32);
	volatile otto_uword32 *ram = od->ram;
	otto_uword32 prime1 = 32771;
	otto_uword32 prime2 = 2147483659U;
	otto_uword32 invert;

	for (invert = 0; invert != 2; invert++) {
		otto_uword32 t = 0;
		unsigned addr = 0;
		for (i = 0; i != n; i++) {
			ram[addr] = t ^ (-invert);
			addr += prime1;
			t += prime2;
			addr %= n;
		}
		t = 0;
		addr = 0;
		for (i = 0; i != n; i++) {
			if (ram[addr] != (t ^ (-invert))) {
				printf ("otto: RAM location %x failed: wrote %x read %x\n",
					addr, t ^ (-invert), ram[addr]);
				return (-1);
			}
			addr += prime1;
			t += prime2;
			addr %= n;
		}
	}

	/* zero the RAM */
	for (i = 0; i != n; i++) {
		ram[i] = 0;
	}

	return (0);
}

/* tell the hardware where the report ring is */
/* set tag bits in ring to make them all invalid */
void otto_setreportbase (struct otto_data *od)
{
	unsigned long phys = OTTO_SVTOPHY ((char *) od->reportbase);
	otto_uword32 addr = OTTO_PHYS_TO_RPTBASE (phys);
	int i;

	for (i = 0; i != OTTO_REP_RING_LEN_REPS; i++) {
		od->reportbase[i].dmareport = OTTO_REP_DMA_TAG;
	}
	od->nextreport = 0;
	od->reporttag = OTTO_REP_DMA_NEGTAG;

	od->dma->report = addr;
	wbflush ();
}

/* build the cell free lists */
void otto_buildcelllists (struct otto_data *od,
        unsigned txcells, unsigned noackcells)
{
	unsigned rxcells;		/* rxcells = # of rx cells */
	unsigned i;
	unsigned freestart;
	volatile otto_uword32 *cells = &od->ram[OTTO_CELLSPOS];

	/* make txcells be a multiple of the VRAM row size */
	i = txcells % OTTO_CELLS_PER_ROW;
	if (i != 0) {
		txcells += OTTO_CELLS_PER_ROW - i;
	}

	rxcells = od->ncellbuffers - txcells - OTTO_UNUSABLE_CELLS;

	od->rxcells = rxcells;
	od->txcells = txcells;
	od->noackcells = noackcells;

	if (od->ncellbuffers <= txcells+OTTO_PANIC_CELLS+OTTO_UNUSABLE_CELLS) {
		panic("otto_buildcelllists: too many transmit calls requested");
	}

	if (noackcells + OTTO_PANIC_CELLS > rxcells) {
		panic ("otto_buildcelllists: too many receive cells requested");
	}


	/* receive cells will be in RAM addresses
		OTTO_UNUSABLE_CELLS..rxcells+OTTO_UNUSABLE_CELLS-1
	   transmit cells will be in RAM addresses
	   	rxcells+OTTO_UNUSABLE_CELLS..od->ncellbuffers

	/* give first two receive cells to hardware directly */
	i = OTTO_UNUSABLE_CELLS;
	od->ram[0] = i++;
	wbflush ();
	*od->exec = OTTOEXEC_cell_wr_rxbuf;
	wbflush ();

	od->ram[0] = i++;
	wbflush ();
	*od->exec = OTTOEXEC_cell_wr_rxbuf;
	wbflush ();

	freestart = i;
	/* make receive cell list */
	while (i != rxcells+OTTO_UNUSABLE_CELLS) {
		cells[i] = OTTO_MAKE_FREE_RXCELL (i+1);
		i++;
	}
	cells[rxcells+OTTO_UNUSABLE_CELLS-1] = OTTO_MAKE_TAIL_FREE_RXCELL;

	/* write list pointers to hardware */
	od->ram[0] = rxcells+OTTO_UNUSABLE_CELLS-1;
	wbflush ();
	*od->exec = OTTOEXEC_load_rxbuftail;
	wbflush ();

	od->ram[0] = freestart;
	wbflush ();
	*od->exec = OTTOEXEC_load_rxbufhead;
	wbflush ();

	od->ram[0] = 0;
	wbflush ();
	*od->exec = OTTOEXEC_load_rxbufpdma;
	wbflush ();

	od->ram[0] = ((rxcells-OTTO_PANIC_CELLS) << 16) | (rxcells - noackcells);
	wbflush ();
	*od->exec = OTTOEXEC_load_rxbufctr;
	wbflush ();
	od->ram[0] = 0;  /* cooked, VC 0 for OAM */
	wbflush ();
	*od->exec = OTTOEXEC_load_recvmode;
	wbflush ();

	/* make transmit cells */
	for (i = rxcells+OTTO_UNUSABLE_CELLS; i != od->ncellbuffers; i++) {
		cells[i] = OTTO_MAKE_FREE_TXCELL (i+1);
	}
	cells[od->ncellbuffers-1] = OTTO_MAKE_TAIL_FREE_TXCELL;

	od->ram[0] = rxcells+OTTO_UNUSABLE_CELLS;
	wbflush ();
	*od->exec = OTTOEXEC_load_txbufhead;
	wbflush ();
	od->ram[0] = od->ncellbuffers-1;
	wbflush ();
	*od->exec = OTTOEXEC_load_txbuftail;
	wbflush ();
	od->ram[0] = txcells - 3; /* XXX should be -1 --- not yet debugged */
	wbflush ();
	*od->exec = OTTOEXEC_load_txbufctr;
	wbflush ();

	*od->exec = OTTOEXEC_read_rxlowbuf;  /* Initialize low water mark. */
	wbflush ();
}

/* initialize the SUNI chip */
void otto_initsuni (struct otto_data *od)
{
	od->suni[8*SUNI_MRI] = 0;	/* take SUNI out of reset */
	od->suni[8*SUNI_RACPMHM] = 0xff;  /* Take cells if hdr ~0 */
	if (od->otto_sdh) {
		od->suni[8*SUNI_TPOPAP1] = 0x98;     /* Enable SDH */
	} else {
		od->suni[8*SUNI_TPOPAP1] = 0x90;     /* Enable SONET */
	}
	od->suni[8*SUNI_RACPCS] = SUNI_RACPCS_DISCOR|SUNI_RACPCS_HCSADD;
			/* no header error correction*/
	wbflush();
}

/* Put suni chip into reset state */
void otto_resetsuni (struct otto_data *od)
{
	od->suni[8*SUNI_MRI] = 0x80;
}

/* allocate a ring */
unsigned otto_allocring (struct otto_data *od)
{
	unsigned x = od->ringfreelist;
	if (x != 0) {
		od->ringfreelist = od->rings[x].next;
	}
	return (x);
}

/* free a ring */
void otto_freering (struct otto_data *od, unsigned index)
{
	od->rings[index].next = od->ringfreelist;
	od->ringfreelist = index;
}

/* allocate a receive packet */
unsigned otto_allocrxpkt (struct otto_data *od)
{
	unsigned x = od->rxpktfreelist;
	if (x != 0) {
		od->rxpktfreelist = od->rxpkts[x].next;
	}
	return (x);
}

/* free a receive packet */
void otto_freerxpkt (struct otto_data *od, unsigned index)
{
	od->rxpkts[index].next = od->rxpktfreelist;
	od->rxpkts[index].m = (struct mbuf *) 0;
	od->rxpktfreelist = index;
}

/* allocate a transmit packet */
unsigned otto_alloctxpkt (struct otto_data *od)
{
	unsigned x = od->txpktfreelist;
	if (x != 0) {
		od->txpktfreelist = od->txpkts[x].next;
	}
	return (x);
}

/* free a transmit packet */
void otto_freetxpkt (struct otto_data *od, unsigned index)
{
	od->txpkts[index].next = od->txpktfreelist;
	od->txpkts[index].m = (struct mbuf *) 0;
	od->txpktfreelist = index;
}

/* Set the ign, pkt, ack and ring values in a receive VC's structure.
   For each of ign, pkt, ack and ring, the value -1 indicates no change. */
void otto_setrxvc (struct otto_data *od, unsigned vc,
        int ign, int pkt, int ack, int ring)
{
	volatile otto_uword32 *rxvc_ring = od->ram_writelo + 
				OTTO_VCPOS+(1*OTTO_MAXVC);
	volatile otto_uword32 *rxvc_tail = od->ram_writelo + 
				OTTO_VCPOS+(2*OTTO_MAXVC);
	otto_uword32 x;
	volatile otto_uword32 *rxvc_head = od->ram_writehi + 
				OTTO_VCPOS+(2*OTTO_MAXVC);

	if (ring != -1) {
		x = (OTTO_MASK_RXVCRING << 16) | OTTO_MAKE_RXVCRING (ring);
		rxvc_ring[vc] = x;
	}
	if (pkt != -1 || ack != -1) {
		x = 0;

		if (pkt != -1) {
			x |= (OTTO_RXVC_PKT << 16) | (pkt? OTTO_RXVC_PKT: 0);
		}
		if (ack != -1) {
			x |= (OTTO_RXVC_ACK << 16) | (ack? OTTO_RXVC_ACK: 0);
		}
		rxvc_tail[vc] = x;
	}
	if (ign != -1) {
		x = (OTTO_RXVC_IGNH << 16) | (ign? OTTO_RXVC_IGNH: 0);
		rxvc_head[vc] = x;
	}
}

void otto_printvc (struct otto_data *od, unsigned vc)
{
	volatile otto_uword32 *txvc_tail = od->ram + 
				OTTO_VCPOS+(0*OTTO_MAXVC);
	volatile otto_uword32 *txvc_cred = od->ram + 
				OTTO_VCPOS+(1*OTTO_MAXVC);
	volatile otto_uword32 *rxvc_tail = od->ram + 
				OTTO_VCPOS+(2*OTTO_MAXVC);
	printf ("vcget: vc%d %x %x %x\n", vc, txvc_cred[vc], txvc_tail[vc],
		rxvc_tail[vc]);
}

/* Set the usesched, enflowcntl, overflow and credits values in a transmit
   VC's structure.
   For each of usesched, enflowcntl, overflow and credits, the value -1
   indicates no change. */
void otto_settxvc (struct otto_data *od, unsigned vc,
        int usesched, int enflowcntl, int overflow, int credits)
{
	volatile otto_uword32 *txvc_cred = od->ram_writehi + 
				OTTO_VCPOS+(1*OTTO_MAXVC);
	volatile otto_uword32 *txvc_tail = od->ram_writelo + 
				OTTO_VCPOS+(0*OTTO_MAXVC);
	otto_uword32 x;

	if (enflowcntl != -1 || overflow != -1 || credits != -1) {
		x = 0;

		if (enflowcntl != -1) {
			x |= (OTTO_TXVC_ENFLOWCNTL << 16) |
				(enflowcntl? OTTO_TXVC_ENFLOWCNTL: 0);
		}
		if (overflow != -1) {
			x |= (OTTO_TXVC_OVERFLOW << 16) |
				(overflow? OTTO_TXVC_OVERFLOW: 0);
		}
		if (credits != -1) {
			x |= (OTTO_MASK_TXVCCREDITS << 16) |
				OTTO_MAKE_TXVCCREDITS (credits);
			x |= (OTTO_TXVC_NOCREDIT << 16) | 
				((credits & 0x1000)? OTTO_TXVC_NOCREDIT: 0);
		}
		txvc_cred[vc] = x;
	}
	if (usesched != -1) {
		x = (OTTO_TXVC_USESCHED << 16) |
			(usesched? OTTO_TXVC_USESCHED: 0);
		txvc_tail[vc] = x;
	}
	od->activate[vc] = OTTO_ACTIVATE_ONLY;
}

/* Sets schedule slot index to indicate the given vc.
   A vc of zero means no vc should use this slot. */
void otto_setsched (struct otto_data *od, unsigned index, unsigned vc)
{
	volatile otto_uword32 *sched = od->ram_writehi + OTTO_SCHEDPOS;

	if (vc == 0) {
		sched[index] = (0xffff0000 | 0);	/* empty slot */
	} else {
		sched[index] = (0xffff0000 | OTTO_MAKE_SCHED (vc));
	}
}

/* waits a while for the SUNI and PLL to acquire lock */
/* It's unclear what happens if we continue operating the device 
  when lock isn't acquired.   However, since we may need to call this
  from an interrupt routine, it waits only a limited time. */
void otto_wait_for_phaselock (struct otto_data *od)
{
	int i;
	int bounce;
	otto_uword32 racp;

	od->suni[8*SUNI_RACPCS] =
		SUNI_RACPCS_FIFORST|SUNI_RACPCS_HCSADD|SUNI_RACPCS_DISCOR;
					 /* reset recv FIFO */
	wbflush ();
	od->suni[8*SUNI_RACPMHM] = 0xff;  /* Take cells if hdr ~0 */
	wbflush ();

	i = 0;
	do {			/* Wait for cell sync */
		racp = od->suni[8*0x50] & 0xff;
		i++;
	} while ((racp & 0x80) != 0 && i != 1000000);

	bounce = 0;
	for (i = 0; i != 5000; i++) {       /* Wait for bouncing to calm down */
		racp = od->suni[8*SUNI_RACPCS] & 0xff;
		if ((racp & 0x80) != 0) {
			bounce++;
		}
	}

	racp = od->suni[8*SUNI_RACPIES] & 0xff; /* Gobble up RACP.OOCDI bit */
	od->suni[8*SUNI_RACPCHCS] = 0;          /* transfer HCS counters */
	od->suni[8*SUNI_RACPCS] = SUNI_RACPCS_HCSADD|SUNI_RACPCS_DISCOR;   
				/* unreset FIFO */
	wbflush ();
}


#define TABLE_LEN  (OTTO_SCHEDSIZE / 2)

/* install a new schedule in the hardware from the softc */
void otto_newsched (struct otto_data *od)
{
	int curr_table;
	int start_index;
	int end_index;
	int i;
	OTTO_SMP_LOCKTMP s0;
	OTTO_SMP_LOCKTMP s1;

	OTTO_SMP_LOCK (&od->schedlock, s1);

	/* find curr_table (0 or 1) */
	for (;;) {
		OTTO_SMP_LOCK (&od->execlock, s0);
		*od->exec = OTTOEXEC_sched_read;
		wbflush ();
		curr_table = (od->ram[0] >> 12) & 1;
		wbflush ();
		OTTO_SMP_UNLOCK (&od->execlock, s0);
		if (curr_table == od->schedhalf) {
			break;
		}
		OTTO_SMP_UNLOCK (&od->schedlock, s1);
		sleep (&lbolt, PZERO + 1);
		OTTO_SMP_LOCK (&od->schedlock, s1);
	}

	od->schedhalf = 1 - curr_table;
	start_index = (od->schedhalf * TABLE_LEN) + TABLE_LEN - od->schedlen;
	end_index = (od->schedhalf * TABLE_LEN) + TABLE_LEN;

	for (i = start_index; i < end_index; i++) {
		otto_setsched (od, i, od->sched[i - start_index]);
	}

	od->ram[OTTO_SCHEDULESTART] = start_index;
	OTTO_SMP_UNLOCK (&od->schedlock, s1);
}


/* Cause "vc"  to have "bw" slots allocated in the schedule.
   Called with softc lock held. */
int otto_update_bandwidth (struct otto_data *od, unsigned vc, int bw)
{
	struct otto_vcdata *vcdata = &od->vcs[vc];
	int change = 0;

	if (bw != vcdata->bandwidth) {
		if (bw >= 0 && bw <= vcdata->bandwidth + od->schedfree) {
			unsigned i;
			unsigned x = od->schedlen;
			unsigned offset = (od->schedhalf * TABLE_LEN) +
				TABLE_LEN - od->schedlen;

			change = 1;
			if (vcdata->bandwidth != 0) {
				od->schedfree += vcdata->bandwidth;
				for (i = 0; i != x; i++) {
					if (od->sched[i] == vc) {
						od->sched[i] = 0;
						otto_setsched (od, i+offset, 0);
					}
				}
			}
			od->schedfree -= bw;
			vcdata->bandwidth = bw;
			if (bw != 0) {
				unsigned every = x / bw;
				while (bw != 0 && every != 0) {
					i = 0;
					while (i < x && bw != 0) {
						if (od->sched[i] == 0) {
							od->sched[i] = vc;
							otto_setsched (od,
								i+offset, vc);
							i += every;
							bw--;
						} else {
							i++;
						}
					}
					every >>= 1;
				}
				if (bw != 0) {
					panic ("otto: unable to set schedule");
				}
			}
		} else {
			change = -1;
		}
	}
	return (change);
}

/* Below is a tranlation table from OTTO ack fifo lfsr values to binary.
   It contains the translations only for lfsr values from 0 to some limit.
   It is used by stepping the lfsr on in software until such a low value is
   reached, then doing the lookup and subtraction mod 8191.
   This allows the lookup table to be much smaller, and doesn't take
   much time on average.

   The procedure lfsr_to_binary does this operation. */
static unsigned short ackfifo_lfsr[] = {
	0x0000, 0x0c6b, 0x0c28, 0x16b6, 0x1293, 0x0963, 0x1dbb, 0x1ec0, 
	0x0655, 0x0f14, 0x0c57, 0x123f, 0x09b8, 0x0271, 0x0cfa, 0x1628, 
	0x1fa7, 0x06cf, 0x17b2, 0x002e, 0x1e4b, 0x1499, 0x0f4e, 0x1fac, 
	0x0aaa, 0x0690, 0x101c, 0x19b4, 0x0c7e, 0x157c, 0x1e8b, 0x04d9, 
	0x014c, 0x0587, 0x176b, 0x0c39, 0x0f88, 0x1b29, 0x1194, 0x127d, 
	0x05a4, 0x0d47, 0x0b64, 0x144d, 0x1401, 0x0527, 0x1c39, 0x0bd6, 
	0x142b, 0x1818, 0x1a47, 0x156d, 0x0fda, 0x03cf, 0x0012, 0x06d4, 
	0x16e6, 0x1b45, 0x1133, 0x1950, 0x0c13, 0x0248, 0x191a, 0x1f22, 
	0x05de, 0x0468, 0x1be8, 0x1c70, 0x19f2, 0x08e7, 0x0097, 0x1fc3, 
	0x0bd0, 0x037e, 0x1960, 0x07ad, 0x09e3, 0x0171, 0x178a, 0x1333, 
	0x0e6a, 0x0305, 0x022a, 0x012d, 0x0a59, 0x0f61, 0x18e8, 0x17b7, 
	0x1a99, 0x1153, 0x0261, 0x08a6, 0x0797, 0x05b3, 0x065c, 0x1665, 
	0x1cc6, 0x11f4, 0x06f7, 0x1efc, 0x165e, 0x1102, 0x1e00, 0x0506, 
	0x0b19, 0x1b75, 0x066d, 0x187d, 0x0cd1, 0x16d2, 0x1b97, 0x0384, 
	0x08bb, 0x131a, 0x070e, 0x0333, 0x189d, 0x0ee1, 0x0877, 0x0033, 
	0x184c, 0x0411, 0x0ace, 0x0db7, 0x1a64, 0x1fe7, 0x1501, 0x1c4b, 
	0x1edd, 0x0299, 0x198e, 0x16a6, 0x1a26, 0x07cd, 0x11b5, 0x0866, 
	0x0a93, 0x10df, 0x0e48, 0x1040, 0x090b, 0x0478, 0x0be9, 0x0932, 
	0x1c33, 0x1b91, 0x0a28, 0x1590, 0x049b, 0x0486, 0x0d74, 0x1e50, 
	0x1696, 0x1370, 0x0063, 0x12ac, 0x07df, 0x0972, 0x02e0, 0x1216, 
	0x123b, 0x19b0, 0x1449, 0x194c, 0x07a9, 0x08a2, 0x1879, 0x0db3, 
	0x103c, 0x12a8, 0x0ae2, 0x0ae6, 0x019d, 0x1f0d, 0x0a2e, 0x1966, 
	0x010f, 0x07f8, 0x1526, 0x11cc, 0x0caa, 0x0d82, 0x14eb, 0x149e, 
	0x1d1d, 0x173d, 0x07e7, 0x0aea, 0x05ac, 0x1e93, 0x19fa, 0x0716, 
	0x0a09, 0x1640, 0x13ba, 0x04bc, 0x031f, 0x134f, 0x1735, 0x16f9, 
	0x1ad5, 0x1c19, 0x015c, 0x01a1, 0x0ed3, 0x1802, 0x062c, 0x1b16, 
	0x0790, 0x004d, 0x0eb5, 0x1477, 0x1e68, 0x0af8, 0x12e3, 0x0f53, 
	0x0cbe, 0x0ce8, 0x1708, 0x1f11, 0x0054, 0x1e9a, 0x0f1b, 0x1109, 
	0x05f4, 0x148b, 0x19e2, 0x108a, 0x1f33, 0x1481, 0x1d52, 0x1adf, 
	0x1bba, 0x10e9, 0x100d, 0x0a32, 0x06a7, 0x0720, 0x1596, 0x07b3, 
	0x1d15, 0x0252, 0x1d3d, 0x119e, 0x15f8, 0x1249, 0x106d, 0x1fb1, 
	0x1d63, 0x030f, 0x03c1, 0x196a, 0x0817, 0x0510, 0x1080, 0x1a6e, 
	0x0555, 0x15a3, 0x0e23, 0x0d92, 0x11e7, 0x1cff, 0x01c3, 0x12fd, 
	0x10ce, 0x0ea7, 0x1e23, 0x0113, 0x114c, 0x0ce1, 0x0df5, 0x17f8, 
	0x1d83, 0x1256, 0x1d2f, 0x18c7, 0x1b0c, 0x1b53, 0x1f89, 0x0aaf, 
	0x01d7, 0x0890, 0x11c4, 0x07fc, 0x1c84, 0x1674, 0x1837, 0x130d, 
	0x13fb, 0x0ccb, 0x0197, 0x06a1, 0x1518, 0x0f9c, 0x1462, 0x0213, 
	0x0bff, 0x02f8, 0x1206, 0x152a, 0x151e, 0x007d, 0x04a1, 0x09e9, 
	0x00e9, 0x04c9, 0x045b, 0x15af, 0x1afd, 0x1779, 0x1f71, 0x0695, 
	0x05c6, 0x1c0f, 0x01ea, 0x11d0, 0x1dd3, 0x12c8, 0x17d8, 0x0a1a, 
	0x0107, 0x199b, 0x0997, 0x0a4a, 0x1e78, 0x06c2, 0x0b33, 0x03f7, 
	0x1a39, 0x083e, 0x0281, 0x0cae, 0x13df, 0x0b4c, 0x04b2, 0x094f, 
	0x1a89, 0x137d, 0x04ef, 0x17a4, 0x09f8, 0x1636, 0x034b, 0x1021, 
	0x1e3c, 0x0dfc, 0x0981, 0x0d86, 0x0ebc, 0x1a01, 0x0c5e, 0x1e07, 
	0x1a92, 0x0cb7, 0x1e35, 0x1345, 0x1ddc, 0x021c, 0x1e2c, 0x125f, 
	0x057a, 0x1eaf, 0x08f8, 0x14ef, 0x0fa2, 0x18bb, 0x048c, 0x0177, 
	0x05fd, 0x1973, 0x12ec, 0x180b, 0x0914, 0x1e59, 0x07f0, 0x19b9, 
	0x16ef, 0x0c42, 0x0d03, 0x14a2, 0x0a62, 0x133c, 0x0700, 0x041a, 
	0x1f45, 0x001f, 0x1388, 0x0a3e, 0x1ab6, 0x0360, 0x028c, 0x02bc, 
	0x0571, 0x18d8, 0x126a, 0x1d21, 0x0433, 0x0043, 0x0a6d, 0x08d2, 
	0x044d, 0x0534, 0x12d3, 0x1ea6, 0x0e32, 0x0f6f, 0x146d, 0x0c83, 
	0x0ca2, 0x15c8, 0x011e, 0x1741, 0x1616, 0x1b84, 0x1d8e, 0x1c94, 
	0x0f49, 0x000d, 0x18e3, 0x0872, 0x0d6f, 0x14e6, 0x12de, 0x1068, 
	0x1f84, 0x1f6c, 0x0346, 0x07eb, 0x1468, 0x185d, 0x0d7a, 0x1790, 
	0x19a3, 0x19c1, 0x0085, 0x0804, 0x15d0, 0x1798, 0x1862, 0x1581, 
	0x025a, 0x1701, 0x097a, 0x0aee, 0x0d4f, 0x04e1, 0x08ef, 0x033b, 
	0x0960, 0x026e, 0x1496, 0x1579, 0x1b26, 0x0524, 0x03cc, 0x0245, 
	0x08e4, 0x016e, 0x0f5e, 0x05b0, 0x10ff, 0x16cf, 0x0ede, 0x1fe4, 
	0x07ca, 0x0475, 0x0483, 0x096f, 0x089f, 0x1f0a, 0x0d7f, 0x1e90, 
	0x134c, 0x17ff, 0x0af5, 0x1e97, 0x147e, 0x071d, 0x1246, 0x050d, 
	0x1cfc, 0x0cde, 0x1b50, 0x1671, 0x0f99, 0x007a, 0x1776, 0x12c5, 
	0x06bf, 0x0b49, 0x1633, 0x19fe, 0x0219, 0x18b8, 0x1e56, 0x1339, 
	0x035d, 0x0040, 0x0f6c, 0x1b81, 0x14e3, 0x185a, 0x1795, 0x04de, 
	0x0521, 0x16cc, 0x1f07, 0x071a, 0x0077, 0x18b5, 0x1857, 0x18b2, 
	0x0e63, 0x0789, 0x1a82, 0x07c3, 0x10b8, 0x1f48, 0x098e, 0x1e18, 
	0x0f34, 0x0756, 0x0a84, 0x1ee1, 0x1187, 0x0022, 0x13cc, 0x02ce, 
	0x14cf, 0x1ac6, 0x1098, 0x0e0b, 0x1f54, 0x138b, 0x19a8, 0x0005, 
	0x085c, 0x1cec, 0x1a56, 0x029d, 0x0dea, 0x0a41, 0x0dde, 0x1a11, 
	0x1233, 0x0f41, 0x00b0, 0x0feb, 0x1cd9, 0x1ab9, 0x0faf, 0x0bc1, 
	0x117b, 0x1bd4, 0x01b8, 0x1992, 0x0c05, 0x0363, 0x169c, 0x05e4, 
	0x00fe, 0x0e18, 0x1d71, 0x10ac, 0x1cb6, 0x028f, 0x19c6, 0x0c70, 
	0x0ef6, 0x0bb1, 0x1dae, 0x16aa, 0x0646, 0x02bf, 0x077c, 0x0731, 
	0x059e, 0x0b13, 0x1036, 0x1bb4, 0x0bf9, 0x0574, 0x1f7e, 0x06b9, 
	0x1175, 0x18fc, 0x17c7, 0x1a2a, 0x1441, 0x18db, 0x0da3, 0x0fbe, 
	0x10bf, 0x1909, 0x0b06, 0x09a6, 0x1750, 0x126d, 0x008a, 0x0c2d, 
	0x1de3, 0x10d5, 0x0d56, 0x07d1, 0x1e6f, 0x1d24, 0x09bf, 0x1ccd, 
	0x1df3, 0x1561, 0x0d2d, 0x0203, 0x1554, 0x0436, 0x0562, 0x076b, 
	0x1902, 0x1caa, 0x0b2a, 0x11b9, 0x02fe, 0x0046, 0x1376, 0x046e, 
	0x182c, 0x03ee, 0x063a, 0x0b99, 0x0928, 0x0a70, 0x0809, 0x16bb, 
	0x1058, 0x171e, 0x1944, 0x086a, 0x03ae, 0x08d5, 0x060f, 0x03e1, 
	0x0f00, 0x1366, 0x0ff5, 0x1608, 0x13d6, 0x0450, 0x1395, 0x1548, 
	0x175a, 0x17e9, 0x0fc8, 0x0a97, 0x0d37, 0x0537, 0x1728, 0x1ef0, 
	0x1bce, 0x1ca4, 0x0940, 0x0186, 0x1871, 0x12d6, 0x15d5, 0x1298, 
	0x0d41, 0x1b6f, 0x12a2, 0x10e3, 0x02f2, 0x1ea9, 0x1f66, 0x0b43, 
	0x00b8, 0x1f5c, 0x0dab, 0x1060, 0x1fd5, 0x0e35, 0x15dd, 0x00c0, 
	0x17cd, 0x0946, 0x03a2, 0x0e4c, 0x120c, 0x0f72, 0x0069, 0x1bee, 
	0x099f, 0x091c, 0x04a9, 0x01df, 0x0126, 0x1470, 0x179d, 0x0968, 
	0x1d45, 0x0327, 0x02e8, 0x1044, 0x0b6c, 0x0c86, 0x009f, 0x08c3, 
	0x1e46, 0x0fd5, 0x0a54, 0x1898, 0x0496, 0x0ca5, 0x1e63, 0x15f3, 
	0x1b07, 0x1af8, 0x09f3, 0x090f, 0x0e2d, 0x15cb, 0x089a, 0x14de, 
	0x1f4f, 0x1cb1, 0x174b, 0x0923, 0x186c, 0x0121, 0x1867, 0x1dc0, 
	0x0223, 0x0eae, 0x04e8, 0x047c, 0x0aff, 0x1744, 0x0278, 0x11fb, 
	0x1bc4, 0x0b79, 0x11a8, 0x1c64, 0x0f25, 0x1619, 0x1359, 0x1417, 
	0x1a30, 0x018c, 0x1220, 0x0bed, 0x1530, 0x1b87, 0x12b2, 0x1c76, 
	0x1435, 0x00cd, 0x1457, 0x0835, 0x1dc5, 0x1d91, 0x1586, 0x1ec5, 
	0x07a1, 0x0d67, 0x1fcd, 0x0936, 0x0677, 0x1c97, 0x1324, 0x1166, 
	0x0c26, 0x1db9, 0x0c55, 0x0cf8, 0x17b0, 0x0f4c, 0x101a, 0x1e89, 
	0x1769, 0x1192, 0x0b62, 0x1c37, 0x1a45, 0x0010, 0x1131, 0x1918, 
	0x1be6, 0x0095, 0x195e, 0x1788, 0x0228, 0x18e6, 0x025f, 0x065a, 
	0x06f5, 0x1dfe, 0x066b, 0x1b95, 0x070c, 0x0875, 0x0acc, 0x14ff, 
	0x198c, 0x11b3, 0x0e46, 0x0be7, 0x0a26, 0x0d72, 0x0061, 0x02de, 
	0x1447, 0x1877, 0x0ae0, 0x0a2c, 0x1524, 0x14e9, 0x07e5, 0x19f8, 
	0x13b8, 0x1733, 0x015a, 0x062a, 0x0eb3, 0x12e1, 0x1706, 0x0f19, 
	0x19e0, 0x1d50, 0x100b, 0x1594, 0x1d3b, 0x106b, 0x03bf, 0x107e, 
	0x0e21, 0x01c1, 0x1e21, 0x0df3, 0x1d2d, 0x1f87, 0x11c2, 0x1835, 
	0x0195, 0x1460, 0x1204, 0x049f, 0x0459, 0x1f6f, 0x01e8, 0x17d6, 
	0x0995, 0x0b31, 0x027f, 0x04b0, 0x04ed, 0x0349, 0x097f, 0x0c5c, 
	0x1e33, 0x1e2a, 0x08f6, 0x048a, 0x12ea, 0x07ee, 0x0d01, 0x06fe, 
	0x1386, 0x028a, 0x1268, 0x0a6b, 0x12d1, 0x146b, 0x011c, 0x1d8c, 
	0x18e1, 0x12dc, 0x0344, 0x0d78, 0x0083, 0x1860, 0x0978, 0x08ed, 
	0x1494, 0x03ca, 0x0f5c, 0x0edc, 0x0481, 0x0d7d, 0x0af3, 0x1244, 
	0x1b4e, 0x1774, 0x1631, 0x1e54, 0x0f6a, 0x1793, 0x1f05, 0x1855, 
	0x1a80, 0x098c, 0x0a82, 0x13ca, 0x1096, 0x19a6, 0x1a54, 0x0ddc, 
	0x00ae, 0x0fad, 0x01b6, 0x169a, 0x1d6f, 0x19c4, 0x1dac, 0x077a, 
	0x1034, 0x1f7c, 0x17c5, 0x0da1, 0x0b04, 0x0088, 0x0d54, 0x09bd, 
	0x0d2b, 0x0560, 0x0b28, 0x1374, 0x0638, 0x0807, 0x1942, 0x060d, 
	0x0ff3, 0x1393, 0x0fc6, 0x1726, 0x093e, 0x15d3, 0x12a0, 0x1f64, 
	0x0da9, 0x15db, 0x03a0, 0x0067, 0x04a7, 0x179b, 0x02e6, 0x009d, 
	0x0a52, 0x1e61, 0x09f1, 0x0898, 0x1749, 0x1865, 0x04e6, 0x0276, 
	0x11a6, 0x1357, 0x121e, 0x12b0, 0x1455, 0x1584, 0x1fcb, 0x1322, 
	0x0c53, 0x1018, 0x0b60, 0x112f, 0x195c, 0x025d, 0x0669, 0x0aca, 
	0x0e44, 0x005f, 0x0ade, 0x07e3, 0x0158, 0x1704, 0x1009, 0x03bd, 
	0x1e1f, 0x11c0, 0x1202, 0x01e6, 0x027d, 0x097d, 0x08f4, 0x0cff, 
	0x1266, 0x011a, 0x0342, 0x0976, 0x0f5a, 0x0af1, 0x162f, 0x1f03, 
	0x0a80, 0x1a52, 0x01b4, 0x1daa, 0x17c3, 0x0d52, 0x0b26, 0x1940, 
	0x0fc4, 0x129e, 0x039e, 0x02e4, 0x09ef, 0x04e4, 0x121c, 0x1fc9, 
	0x0b5e, 0x0667, 0x0adc, 0x1007, 0x1200, 0x08f2, 0x0340, 0x162d, 
	0x01b2, 0x0b24, 0x039c, 0x121a, 0x0ada, 0x033e, 0x039a, 0x0398, 
	0x1fa2, 0x1426, 0x0e65, 0x08b6, 0x1c2e, 0x010a, 0x078b, 0x1d10, 
	0x1d7e, 0x00e4, 0x1a84, 0x05f8, 0x0448, 0x199e, 0x07c5, 0x0358, 
	0x14ca, 0x00f9, 0x10ba, 0x1827, 0x1bc9, 0x099a, 0x1f4a, 0x1430, 
	0x1be1, 0x13b3, 0x0990, 0x148f, 0x102f, 0x0a4d, 0x1e1a, 0x0b59, 
	0x14c5, 0x09d2, 0x0f36, 0x1ba8, 0x14ba, 0x1e7b, 0x0758, 0x0547, 
	0x10c5, 0x1760, 0x0a86, 0x19e6, 0x00ef, 0x06c5, 0x1ee3, 0x1a9f, 
	0x09d7, 0x1689, 0x1189, 0x0e9e, 0x0b7e, 0x0b36, 0x0024, 0x181d, 
	0x1a1e, 0x1aae, 0x13ce, 0x108e, 0x13ed, 0x03fa, 0x02d0, 0x14af, 
	0x13a9, 0x0686, 0x14d1, 0x0d18, 0x1143, 0x1a3c, 0x1ac8, 0x1652, 
	0x09cc, 0x0dce, 0x109a, 0x1f37, 0x0852, 0x0841, 0x0e0d, 0x01f7, 
	0x0f3b, 0x155b, 0x1f56, 0x0b73, 0x11ad, 0x0284, 0x138d, 0x1a4c, 
	0x0581, 0x11ee, 0x19aa, 0x1485, 0x0cc5, 0x0cb1, 0x0007, 0x0cd8, 
	0x1540, 0x0dd4, 0x085e, 0x02b4, 0x1932, 0x13e2, 0x1cee, 0x0e8c
};
/* use the table above to translate from ack fifo lfsr values to binary */
static unsigned lfsr_to_binary (unsigned x)
{
	unsigned y = 0;
	while (x >= (sizeof (ackfifo_lfsr) / sizeof (ackfifo_lfsr[0]))) {
		x = ((x << 1) & 0x1fff) ^ ((x & 0x1000) ? 0 : 0x1601);
		y++;
	}
	return ((ackfifo_lfsr[x] + OTTO_ACKFIFO_MOD - y) % OTTO_ACKFIFO_MOD);
}


/* Atomically read the values of both the read and write pointer of the
   OTTO ACK FIFO.
   The OTTO provides these values in LFSR format, but this routine 
   converts them to binary. */
void otto_ackfifoptrs (struct otto_data *od, unsigned *read, unsigned *write)
{
	otto_uword32 ptrs = *od->ackfifoptrs;

	*read = lfsr_to_binary (OTTO_ACKFIFO_OUT (ptrs));
	*write = lfsr_to_binary (OTTO_ACKFIFO_IN (ptrs));
}

/* return non-zero iff the SUNI is receiving cells and the far end can
   receive our cells (i.e. it's not asserting far-end-receive failure)
   and this state has been true since the last call to this procedure. */
int otto_gettingcells (struct otto_data *od)
{
	unsigned mis = od->suni[8*SUNI_MIS];
	unsigned racpcs = od->suni[8*SUNI_RACPCS];
	unsigned racpies = od->suni[8*SUNI_RACPIES];
	unsigned mc = od->suni[8*SUNI_MC];
	unsigned rsopsis = od->suni[8*SUNI_RSOPSIS] & 0x7f;
	unsigned rlopies = od->suni[8*SUNI_RLOPIES];
	unsigned rlopcs = od->suni[8*SUNI_RLOPCS];
	unsigned rpopsc = od->suni[8*SUNI_RPOPSC];
	unsigned rpopis = od->suni[8*SUNI_RPOPIS] & 0xaf;
	unsigned tacpcs = od->suni[8*SUNI_TACPCS];

	if ((mis & (SUNI_MIS_PFERFI|SUNI_MIS_GPINI|SUNI_MIS_GPINV)) != 0 ||
	    (mc & SUNI_MC_PFERFV) != 0 ||
	    rsopsis != 0 ||
	    (racpcs & SUNI_RACPCS_OOCDV) != 0 ||
            (racpies & (SUNI_RACPIES_OOCDI|SUNI_RACPIES_CHCSI|SUNI_RACPIES_UHCSI|
            	       SUNI_RACPIES_FOVRI|SUNI_RACPIES_FUDRI)) != 0 ||
            (rlopies & (SUNI_RLOPIES_LAISI|SUNI_RLOPIES_LAISI|
                       SUNI_RLOPIES_BIPEI|SUNI_RLOPIES_FEBEI)) != 0 ||
            (rlopcs & (SUNI_RLOPCS_FERFV|SUNI_RLOPCS_LAISV)) != 0 ||
            (rpopsc & (SUNI_RPOPSC_PYEL|SUNI_RPOPSC_PAIS|SUNI_RPOPSC_LOP)) != 0 ||
            rpopis != 0 ||
            (tacpcs & (SUNI_TACPCS_FOVRI|SUNI_TACPCS_TSOCI)) != 0) {
		return (0);
	}
	return (1);
}

/* set LED on link "link" (0 or 1) to:
        on if "state" is 1
        the opposite of its current state if "state" is 2
        off otherwise
 */
void otto_setled (struct otto_data *od, unsigned link, unsigned state)
{
	otto_uword32 x = od->csr->resets;
	otto_uword32 led = link? OTTO_CSR_LEFTLED: OTTO_CSR_RIGHTLED;

	if (state == 2) {
		x ^= led;
	} else if (state == 1) {
		x |= led;
	} else {
		x &= ~led;
	}
	od->csr->leds = x;
	wbflush ();
}

/* Allocate a queue suitable for bandwidth bw, and return it.
   Called with softc lock held */
unsigned otto_allocq (struct otto_data *od, unsigned bw)
{
	unsigned q;
	unsigned bestq = 0;
	unsigned bestcount;
	unsigned bestdiff;

	for (q = 0; q != OTTO_MAXIFQS; q++) {
		unsigned count = od->xmtqueues.per_q[q].count;
		unsigned obw = od->xmtqueues.per_q[q].bw;
		unsigned diff;
		if (q == OTTO_IFQ_NORMAL || q == OTTO_IFQ_SPECIAL) {
			continue;
		}
		if (count == 0) {
			od->xmtqueues.per_q[q].count = 1;
			od->xmtqueues.per_q[q].bw = bw;
			if (bw >= od->schedlen / 2) {
				od->xmtqueues.per_q[q].maxpktswaiting = 2;
			}
			return (q);
		}
		if (obw > bw) {
			diff = obw - bw;
		} else {
			diff = bw - obw;
		}
		if (bestq == 0 ||
		    (count < bestcount && diff <= bestdiff+5) ||
		    (count <= bestcount && diff < bestdiff)) {
			bestq = q;
			bestcount = count;
			bestdiff = diff;
		}
	}
	od->xmtqueues.per_q[bestq].count++;
	return (bestq);
}

/* Deallocate queue q.
   Called with softc lock held */
void otto_deallocq (struct otto_data *od, unsigned q)
{
	if (q != OTTO_IFQ_NORMAL && q != OTTO_IFQ_SPECIAL) {
		unsigned count = od->xmtqueues.per_q[q].count;
		if (count != 0) {
			count--;
			od->xmtqueues.per_q[q].count = count;
		}
		if (count == 0) {
			od->xmtqueues.per_q[q].maxpktswaiting = 1;
		}
	}
}
