/* Access to network interface as a character special device is only
 * so we can use the ioctl() system call to enter the non-standard
 * 48-bit address => PVC mapping.
 *
 * We need an entry in conf.c that has a "struct cdevsw" entry for us that
 * looks like the one at the bottom of the file.
 *
 * We need an entry in /sys/conf/alpha/files like:
 *	io/dec/xxx/thisfile.c optional deviceAbbrev device-driver Binary
 */
#include "otto.h"
#include "otto_platform.h"
#include "sys/uio.h"


#ifndef OTTO_DEBUG
#define OTTO_DEBUG 1
#endif

#if OTTO_DEBUG
#define dputs(str)	printf("%s\n", str)
#define dpanic(str)	panic(str)
/* pass the following a parenthised list: dprintf(("%s\n", s)) */
#define dprintf(list)	printf list
#else
#define dputs(str)	/*nop*/
#define dpanic(str)	/*nop*/
#define dprintf(list)	/*nop*/
#endif


#define minor_to_unit(x)	(x)
#define minor_to_bar(x)		0


static int set_sched (struct otto_data *od, otto_sched_t *sched);
static int get_sched (struct otto_data *od, otto_sched_t *sched);

/*
 * Allow open so that we can ioctl the character special device.
 * Only allow one process at a time at the device, so that use of
 * the special file is not confused.
 */
int otto_open (dev_t dev, int flag)
{
	int unit = minor_to_unit(minor(dev));
	struct otto_softc *sc;

	if (unit >= NOTTO) {
		dputs("Exiting otto_open -- unit out of bounds");
		return (ENXIO);
	}
	sc = &otto_softc[unit];
	if (sc->data.cdevopen) {
		dputs ("Exiting otto_open -- unit already opened");
		return (EBUSY);
	}
	sc->data.cdevopen = 1;
	return (0);
}

int otto_close (dev_t dev, int flag)
{
	int unit = minor_to_unit(minor(dev));
	struct otto_softc *sc;

	if (unit >= NOTTO) {
		dputs("Exiting otto_close -- unit out of bounds");
		return (ENXIO);
	}
	sc = &otto_softc[unit];
	if (! sc->data.cdevopen) {
		dputs("Exiting otto_close -- unit not open");
		return (EIO);
	}
	sc->data.cdevopen = 0;

	return (0);
}

/*
 * use socket-based read/write to do i/o
 *
	otto_read() ["otto_cdev.c"]
	spec_read() ["../../../../src/kernel/vfs/spec_vnops.c":1197]
	ufsspec_read() ["../../../../src/kernel/ufs/ufs_vnops.c":2731]
	vn_read() ["../../../../src/kernel/vfs/vfs_vnops.c":580]
	rwuio() ["../../../../src/kernel/bsd/sys_generic.c":351]
	read() [". ./../../../src/kernel/bsd/sys_generic.c":201]
	syscall() ["../../../../src/kernel/arch/alpha/syscall_trap.c":593]
	_Xsyscall() ["../../../../src/kernel/arch/alpha/locore.s":751]
 */
int otto_read (dev_t dev, struct uio *uio, int flag)
{
	return (ENXIO);
}

int otto_write (dev_t dev, struct uio *uio, int flag)
{
	return (ENXIO);
}


struct otto_oldtablecopy {
	struct otto_map_mapping *table;
	int length;
	int count;
};

/* routine to copy VC map to an old mapping table */
static void *oldcopy (void *arg,
	struct otto_map_newmapping *mapping)
{
	struct otto_oldtablecopy *otc = arg;
	struct otto_map_mapping m;
	int error = 1;
	if  (otc->count < otc->length) {
		m.addr[0] = mapping->addr[0];
		m.addr[1] = mapping->addr[1];
		m.vc = mapping->vc;
		error = copyout ((caddr_t) &m,
			(caddr_t) &otc->table[otc->count++], sizeof (m));
	}
	return ((void*) error);
}

/*
 *  Add things here
 */
int otto_ioctl (dev_t dev, unsigned int cmd, caddr_t data, int flag)
{
	int unit = minor_to_unit (minor(dev));
	struct otto_softc *sc = &otto_softc[unit];
	otto_table_t *table;
	otto_newtable_t *newtable;
	int rc;
	otto_mapping_t *oldmapping;
	otto_newmapping_t *mapping;
	otto_newmapping_t m;
	unsigned ring;
	unsigned vc;
	OTTO_SMP_LOCKTMP s;
	otto_vcsetup_t *vcsetup;
	struct otto_vcdata *vcdata;
	struct otto_oldtablecopy otc;
	unsigned flags;
	int wholepkt = 1;

	if (unit >= NOTTO) {
		dputs("Exiting otto_ioctl -- unit out of bounds");
		return (ENXIO);
	}

	switch (cmd)
	{
	case OTTO_GET_ENTRY:
		/* emulate old call */
		table = (otto_table_t *) data;
		otc.table = table->table;
		otc.length = table->length;
		otc.count = 0;
		/* don't report stuff in portmap because old style
		   clients can't handle it anyway */
		rc = (otto_map_iterate (table->ip?
			&sc->data.ipvcmap :  &sc->data.vcmap,
			oldcopy, &otc) != 0);
		if (rc) {
			return (EFAULT);
		} else {
			table->length = otc.count;
		}
		break;

	case OTTO_GET_ENTRY2:
		newtable = (otto_newtable_t *) data;
		if (newtable->ip) {
			rc = otto_map_copy (&sc->data.ipvcmap,
				newtable->table, newtable->length, 1);
			if (rc != -1) {
				int orc = rc;
				rc = otto_map_copy (&sc->data.portvcmap,
				   newtable->table+orc, newtable->length-orc, 1);
			   	if (rc != -1) {
			   		rc += orc;
			   	}
			}
		} else {
			rc = otto_map_copy (&sc->data.vcmap,
				newtable->table, newtable->length, 1);
		}
		if (rc == -1) {
			return (EFAULT);
		} else {
			newtable->length = rc;
		}
		break;

	case OTTO_DEL_ENTRY:
		/* emulate old call */
		oldmapping = (otto_mapping_t*) data;
		bzero ((char *) m.mapping.addr, sizeof (m.mapping.addr));
		m.mapping.addr[0] = oldmapping->mapping.addr[0];
		m.mapping.addr[1] = oldmapping->mapping.addr[1];
		m.mapping.vc = oldmapping->mapping.vc;
		m.ip = oldmapping->ip;
		mapping = &m;
		goto dodel;
	case OTTO_DEL_ENTRY2:
		mapping = (otto_newmapping_t*) data;
	dodel:
		if (mapping->mapping.vc != 0) {
			if (mapping->mapping.vc >= OTTO_MAXVC) {
				return (EINVAL);
			}
			OTTO_SMP_LOCK (&sc->lk_otto_softc, s);
			flags = sc->data.vcs[mapping->mapping.vc].flags;
			if ((flags & ~OTTO_VC_PVC) & OTTO_VC_ALLOCATED) {
				OTTO_SMP_UNLOCK (&sc->lk_otto_softc, s);
				return (EINVAL);
			}
			sc->data.vcs[mapping->mapping.vc].flags &=
				~(OTTO_VC_PVC|OTTO_VC_ARPXMT|OTTO_VC_ARPRCV|
				  OTTO_VC_ARPRCV_RECENT|OTTO_VC_DONT_ARP);
			sc->data.arpstate.newpvcs = 1;
			OTTO_SMP_UNLOCK (&sc->lk_otto_softc, s);

			while (otto_map_from_vc (&sc->data.portvcmap,
				 mapping->mapping.vc, mapping->mapping.addr)) {
				otto_map_delete (&sc->data.portvcmap,
					    mapping->mapping.addr);
			}
			while (otto_map_from_vc (&sc->data.ipvcmap,
				 mapping->mapping.vc, mapping->mapping.addr)) {
				otto_map_delete (&sc->data.ipvcmap,
					    mapping->mapping.addr);
			}
			while (otto_map_from_vc (&sc->data.vcmap,
				 mapping->mapping.vc, mapping->mapping.addr)) {
				otto_map_delete (&sc->data.vcmap,
					    mapping->mapping.addr);
			}
		} else {
			if (mapping->ip) {
				otto_map_delete (&sc->data.ipvcmap,
					mapping->mapping.addr);
				otto_map_delete (&sc->data.portvcmap,
					mapping->mapping.addr);
			} else {
				otto_map_delete (&sc->data.vcmap,
					mapping->mapping.addr);
			}
		}
		break;

	case OTTO_INS_ENTRY:
		/* emulate old call */
		oldmapping = (otto_mapping_t*) data;
		bzero ((char *) m.mapping.addr, sizeof (m.mapping.addr));
		m.mapping.addr[0] = oldmapping->mapping.addr[0];
		m.mapping.addr[1] = oldmapping->mapping.addr[1];
		m.mapping.vc = oldmapping->mapping.vc;
		m.ip = oldmapping->ip;
		mapping = &m;
		goto doins;
	case OTTO_INS_ENTRY2:
		mapping = (otto_newmapping_t*) data;
	doins:
		if (mapping->mapping.vc==0 || mapping->mapping.vc>=OTTO_MAXVC) {
			return (EINVAL);
		}
		if (mapping->ip) {
			if (mapping->mapping.addr[3] != 0 ||
			    mapping->mapping.addr[5] != 0) {
				sc->data.portmapping = 1;
				otto_map_set (&sc->data.portvcmap,
				    mapping->mapping.addr, mapping->mapping.vc);
			} else {
				otto_map_set (&sc->data.ipvcmap,
				    mapping->mapping.addr, mapping->mapping.vc);
			}
		} else {
			otto_map_set (&sc->data.vcmap,
				mapping->mapping.addr, mapping->mapping.vc);
		}
		break;

	case OTTO_SET_DEFVC:
		vc = *(unsigned *)data;
		if (vc >= OTTO_MAXVC) {
			return (EINVAL);
		}
		OTTO_SMP_LOCK (&sc->lk_otto_softc, s);
		sc->data.defaultvc = vc;
		OTTO_SMP_UNLOCK (&sc->lk_otto_softc, s);
		break;

	case OTTO_GET_DEFVC:
		OTTO_SMP_LOCK (&sc->lk_otto_softc, s);
		*(unsigned *)data = sc->data.defaultvc;
		OTTO_SMP_UNLOCK (&sc->lk_otto_softc, s);
		break;

	case OTTO_SET_VC:
		vcsetup = (otto_vcsetup_t *) data;
		if (vcsetup->vc == 0 || vcsetup->vc >= OTTO_MAXVC) {
			return (EINVAL);
		}
		OTTO_SMP_LOCK (&sc->lk_otto_softc, s);
		vcdata = &sc->data.vcs[vcsetup->vc];

		flags = vcdata->flags;

		if (vcsetup->bandwidth != -1) {
			if (otto_update_bandwidth (
			    &sc->data, vcsetup->vc, vcsetup->bandwidth) == -1) {
				OTTO_SMP_UNLOCK (&sc->lk_otto_softc, s);
				return (EINVAL);
			}
		}
		if ((vcdata->flags & ~OTTO_VC_PVC) & OTTO_VC_ALLOCATED) {
			OTTO_SMP_UNLOCK (&sc->lk_otto_softc, s);
			return (EINVAL);
		}
		if (vcsetup->encap != -1) {
			vcdata->encap = vcsetup->encap & 0xf;
#ifdef CRED_NO_EOP
			if (vcdata->encap == OTTO_EN_CRED)
			  wholepkt = 0;
#endif
		}
		if (vcsetup->etype != -1) {
			vcdata->etype = vcsetup->etype;
		}
		if (vcsetup->ignore != -1) {
			vcdata->flags = (vcdata->flags & ~OTTO_VC_IGNORE) |
				(vcsetup->ignore? OTTO_VC_IGNORE:0);
		}
		if (vcsetup->useflowcontrol != -1) {
			vcdata->flags = (vcdata->flags & ~OTTO_VC_NOFC) |
				(vcsetup->useflowcontrol? 0:OTTO_VC_NOFC);
		}
		if (vcsetup->useschedule != -1) {
			vcdata->flags = (vcdata->flags & ~OTTO_VC_SCHED) |
				(vcsetup->useschedule? OTTO_VC_SCHED:0);
		}
		if (vcdata->q == OTTO_IFQ_SPECIAL) {
			otto_deallocq (&sc->data, vcdata->q);
			vcdata->q = OTTO_IFQ_SPECIAL;
		} else if (vcdata->bandwidth == 0 &&
			(vcdata->flags & OTTO_VC_SCHED) == 0) {
			otto_deallocq (&sc->data, vcdata->q);
			vcdata->q = OTTO_IFQ_NORMAL;
		} else if (vcdata->q == OTTO_IFQ_NORMAL) {
			vcdata->q = otto_allocq (&sc->data, vcdata->bandwidth);
		}
		if (vcsetup->dontarp != -1) {
			vcdata->flags = (vcdata->flags & ~OTTO_VC_DONT_ARP) |
				(vcsetup->dontarp? OTTO_VC_DONT_ARP:0);
		}
#if 0
/* original code here doesn't let us set up FC VC unless the other side of
   the link is FlowMaster-speaking. we'll remove that constraint. Note that
   this lets us set up FC VCs freely, and these may never be able to send
   unless the switch sends us credit for them. */
                if (sc->data.snmp.pkt.fmsupport && vcdata->bandwidth == 0 &&
		    (vcdata->flags & (OTTO_VC_NOFC|OTTO_VC_SCHED))==0) {
#endif
		if (vcdata->bandwidth == 0 &&
		    (vcdata->flags & (OTTO_VC_NOFC|OTTO_VC_SCHED))==0) {
			vcdata->flags |= OTTO_VC_FC;
		} else {
			vcdata->flags &= ~OTTO_VC_FC;
		}
		ring = sc->data.defaultring;
		otto_setrxvc (&sc->data, vcsetup->vc,
		    vcsetup->ignore, wholepkt, -1, sc->data.rings[ring].desc);
		sc->data.vcs[vcsetup->vc].rxring = ring;
		/* always give FC VCs 100 cells of credit to start */
		otto_settxvc (&sc->data, vcsetup->vc,
			vcsetup->useschedule, (vcdata->flags & OTTO_VC_FC),
			-1, ((vcdata->flags & OTTO_VC_FC) ? 100 : -1));
		if ((vcdata->flags & OTTO_VC_PVC) == 0) {
			vcdata->flags |= OTTO_VC_PVC;
			sc->data.arpstate.newpvcs = 1;
		}
		if ((flags & OTTO_VC_FC)==0 && (vcdata->flags &OTTO_VC_FC)!=0) {
		    	/* turning on flow control---do resync */
			fc_resync (sc->data.fcs, vcsetup->vc, vcsetup->vc, 1);
		}
		OTTO_SMP_UNLOCK (&sc->lk_otto_softc, s);
		break;

	case OTTO_GET_VC:
		vcsetup = (otto_vcsetup_t *) data;
		if (vcsetup->vc == 0 || vcsetup->vc >= OTTO_MAXVC) {
			return (EINVAL);
		}
		OTTO_SMP_LOCK (&sc->lk_otto_softc, s);
		vcdata = &sc->data.vcs[vcsetup->vc];
		vcsetup->encap = vcdata->encap;
		vcsetup->etype = vcdata->etype;
		vcsetup->useflowcontrol = (vcdata->flags&OTTO_VC_FC)?1:0;
		vcsetup->ignore = (vcdata->flags&OTTO_VC_IGNORE)?1:0;
		vcsetup->useschedule = (vcdata->flags&OTTO_VC_SCHED)?1:0;
		vcsetup->bandwidth = vcdata->bandwidth;
		vcsetup->dontarp = (vcdata->flags&OTTO_VC_DONT_ARP)?1:0;
		OTTO_SMP_UNLOCK (&sc->lk_otto_softc, s);
		break;

	case OTTO_FLOW_RESYNC:
		OTTO_SMP_LOCK (&sc->lk_otto_softc, s);
		fc_resync (sc->data.fcs, 1, FC_LASTVC, 1);
		OTTO_SMP_UNLOCK (&sc->lk_otto_softc, s);
		break;

	case OTTO_SET_SCHED:
		rc = set_sched (&sc->data, (otto_sched_t *) data);
		if (rc == 0) {
			int i;
			OTTO_SMP_LOCK (&sc->lk_otto_softc, s);
			for (vc = 1; vc != OTTO_MAXVC; vc++) {
				vcdata = &sc->data.vcs[vc];
				if (vcdata->q == OTTO_IFQ_SPECIAL) {
					otto_deallocq (&sc->data, vcdata->q);
					vcdata->q = OTTO_IFQ_SPECIAL;
				} else if (vcdata->bandwidth == 0 &&
					(vcdata->flags & OTTO_VC_SCHED) == 0) {
					otto_deallocq (&sc->data, vcdata->q);
					vcdata->q = OTTO_IFQ_NORMAL;
				} else if (vcdata->q == OTTO_IFQ_NORMAL) {
					vcdata->q = otto_allocq (&sc->data,
						vcdata->bandwidth);
				}
			}
			OTTO_SMP_UNLOCK (&sc->lk_otto_softc, s);
		}
		return (rc);

	case OTTO_GET_SCHED:
		return (get_sched (&sc->data, (otto_sched_t *) data));
	
	case OTTO_SET_FRAMING:
		sc->data.otto_sdh = *((unsigned*) data);
		otto_initsuni (&sc->data);
		break;

	case OTTO_MEASURE:
		OTTO_SMP_LOCK (&sc->lk_otto_softc, s);
		rc = otto_measure_ioctl (&sc->data, (struct otto_measure *)data);
		OTTO_SMP_UNLOCK (&sc->lk_otto_softc, s);
		return (rc);

	default:
		dprintf (("otto_ioctl: Unknown ioctl %x\n", cmd));
		return (EINVAL);
	}
	return (0);
}

#define TABLE_LEN  (OTTO_SCHEDSIZE / 2)

static int set_sched (struct otto_data *od, otto_sched_t *sched)
{
	unsigned table[TABLE_LEN];
	OTTO_SMP_LOCKTMP s;
	unsigned i;
	unsigned vc;
	unsigned len = sched->length;
	int error;

	if (len == 0 || len > TABLE_LEN) {
		return (EINVAL);
	}
	error = copyin ((caddr_t) sched->table, (caddr_t) table,
		    sizeof (table[0]) * len);

	if (error == 0) {
		for (i = 0; i != len; i++) {
			if (table[i] >= OTTO_MAXVC) {
				return (EINVAL);
			}
		}

		OTTO_SMP_LOCK (&od->schedlock, s);
		bcopy ((char *) table, (char *) od->sched,
			sizeof (table[0]) * len);
		od->schedlen = len;
		od->schedfree = 0;
		for (vc = 1; vc != OTTO_MAXVC; vc++) {
			od->vcs[vc].bandwidth = 0;
		}
		for (i = 0; i != len; i++) {
			vc = table[i];
			if (vc != 0) {
				unsigned b = od->vcs[vc].bandwidth;
				/* turn off flow control if VC in schedule */
				if (b==0 && (od->vcs[vc].flags & OTTO_VC_FC)) {
					od->vcs[vc].flags &= ~OTTO_VC_FC;
					otto_settxvc (od, vc, -1, 0, -1, -1);
				}
				od->vcs[vc].bandwidth = b+1;
			} else {
				od->schedfree++;
			}
		}
		OTTO_SMP_UNLOCK (&od->schedlock, s);

		otto_newsched (od);
	}
	return (error);
}

static int get_sched (struct otto_data *od, otto_sched_t *sched)
{
	OTTO_SMP_LOCKTMP s;
	int error = 0;

	OTTO_SMP_LOCK (&od->schedlock, s);
	if (sched->length == 0) {
		/* nothing */
	} else if (sched->length < od->schedlen) {
		error = copyout ((caddr_t) od->sched, (caddr_t) sched->table,
		    sizeof (sched->table[0]) * sched->length);
	} else {
		error = copyout ((caddr_t) od->sched, (caddr_t) sched->table,
		    sizeof (sched->table[0]) * od->schedlen);
	}
	if (error == 0) {
		sched->length = od->schedlen;
	}
	OTTO_SMP_UNLOCK (&od->schedlock, s);
	return (error);
}


int otto_mmap (dev_t dev, off_t off, int prot)
{
	if (off != 0) {
		return (-1);
	}
	if (prot != 1) {
		return (-1);
	}
	if (ottoclockloc == 0) {
		return (-1);
	}
	return (btop (OTTO_SVTOPHY (ottoclockloc)));
}
